#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright 2018 D. de Vries

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This file contains a set of CMDOWS utility functions.
"""
from __future__ import absolute_import, division, print_function

from copy import copy
from lxml.etree import _Element
from typing import Tuple, Union


def get_related_parameter_uid(elem_param, full_xml):
    # type: (_Element, _Element) -> Tuple[str, str]
    """Function to retrieve the UID of the related parameter. This UID refers to the original local in the XML file of
     the parameter and is used when a parameter is found in a CMDOWS file which is an architecture type or a higher
     instance.

    Parameters
    ----------
        elem_param : _Element
            Element of the parameter to be mapped.
        full_xml : _Element
            Element of the full XML file in which the related parameter UID should be found.

    Returns
    -------
        param : str
            Original parameter UID for which the related UID is searched for.
        mapped : str
            The found related parameter UID.
    """
    param = elem_param.attrib['uID']
    if isinstance(elem_param.find('relatedParameterUID'), _Element):
        mapped = elem_param.find('relatedParameterUID').text
    elif isinstance(elem_param.find('relatedInstanceUID'), _Element):
        related_instance = get_element_by_uid(full_xml, elem_param.find('relatedInstanceUID').text)
        mapped = related_instance.find('relatedParameterUID').text
    else:
        raise AssertionError('Could not map element {}.'.format(param))
    return str(param), str(mapped)


def get_element_by_uid(xml, uid):
    # type: (_Element, str) -> _Element
    """Method to get the element based on a UID value.

    Parameters
    ----------
        xml : _Element
            Element of an XML file where the uID should be found.

        uid : str
            uID to be found.

    Returns
    -------
        _Element
            Element with the right UID.
    """
    xpath_expression = get_uid_search_xpath(uid)
    els = xml.xpath(xpath_expression)
    if len(els) > 1:
        raise AssertionError('Multiple elements with UID ' + uid + ' found. Use "check_uids()" to check if all UIDs'
                                                                   ' are unique.')
    elif len(els) == 0:
        raise AssertionError('Could not find element with UID ' + uid + '.')
    return els[0]


def get_uid_search_xpath(uid):
    # type: (str) -> str
    """Method to get the XPath expression for a UID that might contain quote characters.

    Parameters
    ----------
        uid : str
            Original UID string with XPath expression.

    Returns
    -------
        str
            Processed XPath expression to escape quote characters using "concat()".
    """
    if '"' in uid or '&quot;' in uid:
        uid_concat = "concat('%s')" % uid.replace('&quot;', "\',\'\"\',\'").replace('"', "\',\'\"\',\'")
        return './/*[@uID=' + uid_concat + ']'
    else:
        return './/*[@uID="' + uid + '"]'


def get_loop_nesting_obj(elem):
    # type: (_Element) -> Union[list, dict]
    """Function to make an object of the loop hierarchy based on the loopNesting element in a CMDOWS file.

    Parameters
    ----------
        elem : _Element
            Element in the XML file (loopNesting or one of its subelements)

    Returns
    -------
        list or dict
            A list object containing dictionaries and string entries to represent the loopNesting element.

    Raises
    ------
        AssertionError
            If the provided element is not formatted correctly.
    """
    basic_list = []
    basic_dict = {}
    d = copy(elem.attrib)
    if elem.tag == 'loopNesting' or elem.tag == 'loopElements' or elem.tag == 'functionElements':
        for x in elem.iterchildren():
            basic_list.append(get_loop_nesting_obj(x))
        if len(basic_list) == 1 and isinstance(basic_list[0], list):
            return basic_list[0]
        else:
            return basic_list
    elif elem.tag == 'loopElement':
        basic_dict[d['relatedUID']] = []
        for x in elem.iterchildren():
            basic_dict[d['relatedUID']].extend(get_loop_nesting_obj(x))
        return basic_dict
    elif elem.tag == 'functionElement':
        return elem.text
    else:
        raise AssertionError('Something went wrong in this function...')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright 2018 D. de Vries

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This file contains the definition the `DisciplineComponent` class.
"""
from __future__ import absolute_import, division, print_function

from typing import Optional

from .abstract_discipline import AbstractDiscipline
from .xml_component import XMLComponent


class DisciplineComponent(XMLComponent):
    """Specialized `XMLComponent` wrapping an `AbstractDiscipline`.

    This version of `XMLComponent` defines in- and output variables based on the in- and output template XML files
    generated by a subclass of `AbstractDiscipline`. The `execute()` method simply forwards to that of the discipline.

    Attributes
    ----------
        discipline
    """

    def __init__(self, discipline, data_folder='', keep_files=False, base_file=None):
        # type: (AbstractDiscipline, Optional[str]) -> None
        """Initialize a `Component` using a given `discipline`.

        Stores a reference to the given `discipline`. The in- and output XML templates should already exist at the paths
        specified in the `discipline`. This constructor uses those files to create the ``OpenMDAO`` `params` and
        `unknowns` using the methods exposed by the `XMLComponent` class this class inherits from.

        Parameters
        ----------
            discipline : :obj:`AbstractDiscipline`
                Instance of a subclass of `AbstractDiscipline` this `Component` will represent.

            data_folder : str(''), optional
                Path to a folder in which to store (temporary) data of this `Component` during execution.

            keep_files : bool(False), optional
                Set to `True` to keep the data files generated by this `Component` during execution.

            base_file : str, optional
                Path to an XML file which should be kept up-to-date with the latest data, if required.

        Notes
        -----
            Although this constructor could use the supplied `discipline` to also automatically generate its in- and
            output XML templates on the fly, the user is left in control of their generation. This is to allow for a
            `discipline` to generate different in- and output templates dynamically based on certain parameters. During
            execution only the static methods of the `discipline`s are used. Hence, any instance variables will not be
            accessible then. Therefore it is impossible to guarantee consistency if the in- and output XML files are
            generated here.
        """
        self._discipline = discipline
        if discipline.supplies_partials:
            super(DisciplineComponent, self).__init__(self._discipline.in_file,
                                                      self._discipline.out_file,
                                                      self._discipline.partials_file,
                                                      data_folder, keep_files, base_file)
        else:
            super(DisciplineComponent, self).__init__(self._discipline.in_file,
                                                      self._discipline.out_file,
                                                      None,
                                                      data_folder, keep_files, base_file)
            self.partials_from_xml = None

    @property
    def discipline(self):
        # type: () -> AbstractDiscipline
        """:obj:`AbstractDiscipline`: Read-only reference to the specific discipline this `Component` wraps."""
        return self._discipline

    def execute(self, input_xml=None, output_xml=None):
        # type: (str, str) -> None
        """Call the `execute()` method of this `Component`'s discipline.

        Parameters
        ----------
            input_xml : str
                Path to the input XML file.

            output_xml : str
                Path to the output XML file.

        Raises
        ------
            ValueError
                If either no `input_xml` or `output_xml` path was specified.

        Notes
        -----
            Since this class inherits from `XMLComponent` the interface, including the optionality of its arguments, are
            left untouched. For this method this means the `input_xml` and `output_xml` parameters are strictly
            optional. However, in the context of the `DisciplineComponent` they should always be given. Therefore an
            exception is raised here when one of them or both are omitted.
        """
        if input_xml is None or output_xml is None:
            raise ValueError('Both an input_xml and output_xml path are expected.')
        self.discipline.execute(input_xml, output_xml)

    def linearize(self, input_xml=None, partials_xml=None):
        # type: (str, str) -> None
        """Call the `linearize()` method of this `Component`'s discipline.

        Parameters
        ----------
            input_xml : str
                Path to the input XML file.

            partials_xml : str
                Path to the partials XML file.

        Raises
        ------
            ValueError
                If either no `input_xml` or `partials_xml` path was specified.
        """
        if self.discipline.supplies_partials:
            if input_xml is None or partials_xml is None:
                raise ValueError('Both an input_xml and a partials_xml path are expected.')
            self.discipline.linearize(input_xml, partials_xml)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright 2018 D. de Vries

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This file contains the definition the `XMLComponent` class.
"""
from __future__ import absolute_import, division, print_function

import abc
import os
from abc import abstractmethod
from datetime import datetime

import numpy as np
from lxml import etree
from openmdao.api import Group, IndepVarComp, ExplicitComponent
from openmdao.vectors.vector import Vector
from typing import Optional, List, Union, Iterable

from openlego.utils.xml_utils import xml_safe_create_element, xml_to_dict, xpath_to_param, param_to_xpath, xml_merge
from openlego.partials.partials import Partials

dir_path = os.path.dirname(os.path.realpath(__file__))


class XMLComponent(ExplicitComponent):
    """Abstract base class exposing an interface to use XML files for its in- and output.

    This subclass of `PromotingComponent` can automatically create ``OpenMDAO`` inputs and outputs based on given in-
    and output XML template files. For maximum flexibility it is possible to only specify inputs from an XML file and
    retain direct control over the definition of the outputs, or vice versa. It is also perfectly valid to add inputs
    even when an XML file is used to generate a set of inputs, or outputs when an XML file it used to generate outputs.
    It is even possible to generate in- and/or output parameters based on more than one XML file.

    This class exposes the functions `set_inputs_from_xml()` and `set_outputs_from_xml()` for this purpose. Lists of all
    parameters obtained from XML files are stored by this class for later inspection.

    The `solve_nonlinear()` method of the `Component` class is implemented to wrap the XML related operations such as
    reading in- and output data from the corresponding XML files during execution and storing it in this `Component`'s
    parameter dictionaries.

    A new abstract method is defined by this class, `execute()`, which assumes the role of the `solve_nonlinear()`
    function, in essence. A specific case of this class should implement this method to perform the actual calculations
    of an analysis tool using XML in- and/or output.

    Attributes
    ----------
        inputs_from_xml, outputs_from_xml, partials_from_xml : dict
            List of inputs, resp. outputs, resp. partials, taken from XML.

        data_folder : str('')
            Path to a folder in which to store data generated during the execution of this `XMLComponent`.

        keep_files : bool(False)
            Set to `True` to keep all temporary XML files generated by the `XMLComponent` during execution.

            This attribute is `False` by default, in which case all temporary in- and output XML files will be deleted
            after they are no longer needed by this component.

        base_file : str, optional
            Path to an XML file to keep up-to-date with the latest data from executions.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self,
                 input_xml=None,            # type: Optional[Union[str, etree._ElementTree]]
                 output_xml=None,           # type: Optional[Union[str, etree._ElementTree]]
                 partials_xml=None,         # type: Optional[Union[str, etree._ElementTree]]
                 data_folder='',            # type: str
                 keep_files=False,          # type: bool
                 base_file=None             # type: Optional[str]
                 ):
        # type: (...) -> None
        """Initialize the `XMLComponent`.

        Parameters
        ----------
            input_xml, output_xml, partials_xml : str or :obj:`etree._ElementTree`, optional
                Paths to or an `etree._ElementTree` of input, resp. output, resp. partial, XML files.

            data_folder : str
                Path to the folder in which to store temporary data files generated by this `XMLComponent`.

            keep_files : bool(False)
                Set to `True` to keep the temporary XML files after they are no longer needed.

            base_file : str, optional
                Path to a base XML file to keep up-to-date with all latest data from this `XMLComponent`.
        """
        super(XMLComponent, self).__init__()

        self.inputs_from_xml = dict()
        self.outputs_from_xml = dict()
        self.partials_from_xml = dict()

        if input_xml is not None:
            self.set_inputs_from_xml(input_xml)

        if output_xml is not None:
            self.set_outputs_from_xml(output_xml)

        if partials_xml is not None:
            self.declare_partials_from_xml(partials_xml)

        self.data_folder = data_folder
        self.keep_files = keep_files
        self.base_file = base_file

    def set_inputs_from_xml(self, input_xml):
        # type: (Union[str, etree._ElementTree]) -> None
        """Set inputs to the `Component` based on an input XML template file.

        Parameter names correspond to their XML elements' full XPaths, converted to valid ``OpenMDAO`` names using the
        `xpath_to_param()` method.

        Parameters
        ----------
            input_xml : str or :obj:`etree._ElementTree`
                Path to or an `etree._ElementTree` of an input XML file.
        """
        self.inputs_from_xml.clear()
        for xpath, value in xml_to_dict(input_xml).items():
            name = xpath_to_param(xpath)
            self.inputs_from_xml.update({name: value})

    def set_outputs_from_xml(self, output_xml):
        # type: (Union[str, etree._ElementTree]) -> None
        """Set outputs to the `Component` based on an output XML template file.

        Parameter names correspond to their XML elements' full XPaths, converted to valid ``OpenMDAO`` names using the
        `xpath_to_param()` method.

        Parameters
        ----------
            output_xml : str or :obj:`etree._ElementTree`
                Path to or an `etree._ElementTree` of an output XML file.
        """
        self.outputs_from_xml.clear()
        for xpath, value in xml_to_dict(output_xml).items():
            name = xpath_to_param(xpath)
            self.outputs_from_xml.update({name: value})

    def declare_partials_from_xml(self, partial_xml):
        # type: (Union[str, etree._ElementTree]) -> None
        """Declare partials to the `Component` based on a partials XML template file.

        Parameters
        ----------
            partial_xml : str or :obj:`etree._ElementTree`
                Path to or an `etree._ElementTree` of a partials XML file.
        """
        self.partials_from_xml.clear()
        if partial_xml is not None:
            partials = Partials(partial_xml)
            self.partials_from_xml = partials.get_partials().copy()

    @property
    def variables_from_xml(self):
        # type: () -> dict
        """:obj:`dict`: Dictionary of all XML inputs and outputs."""
        variables = self.inputs_from_xml.copy()
        variables.update(self.outputs_from_xml.copy())
        return variables

    def setup(self):
        for name, value in self.inputs_from_xml.items():
            if not isinstance(value, float) and not isinstance(value, np.ndarray):
                # TODO: pass_by_obj
                # raise NotImplementedError('pass-by-object variables are not yet supported by OpenMDAO 2.0')
                pass
            else:
                self.add_input(name, value)

        for name, value in self.outputs_from_xml.items():
            if not isinstance(value, float) and not isinstance(value, np.ndarray):
                # TODO: pass_by_obj
                # raise NotImplementedError('pass-by-object variables are not yet supported by OpenMDAO 2.0')
                pass
            else:
                self.add_output(name, value)

        if self.partials_from_xml:
            for src, partials in self.partials_from_xml.items():
                if src is not None and partials is not None:
                    self.declare_partials(src, partials.keys())
        else:
            self.declare_partials('*', '*', method='fd')
            # if self.outputs_from_xml and self.inputs_from_xml:
            #     for src in self.outputs_from_xml.keys():
            #         self.declare_partials(src, self.inputs_from_xml.keys(), method='fd')

    @abstractmethod
    def execute(self, input_xml=None, output_xml=None):
        # type: (Optional[str], Optional[str]) -> None
        """Execute the tool using the given input XML file. Write the results to the given output XML file.

        Parameters
        ----------
            input_xml, output_xml : str, optional
                Path to the input, resp. output, XML file.
        """
        raise NotImplementedError

    @abstractmethod
    def linearize(self, input_xml=None, partials_xml=None):
        # type: (Optional[str], Optional[str]) -> None
        """Compute the partials of a tool using the given XML file. Write the results to the given partials XML file.

        Parameters
        ----------
            input_xml, partials_xml : str, optional
                Path to the input, resp. partials, XML file.
        """
        raise NotImplementedError

    def generate_file_names(self):
        # type: () -> (str, str, str)
        """Generate temporary file names for the input, output, and partials XML files.

        Returns
        -------
            str
                Input XML file path.

            str
                Output XML file path.

            str
                Partials XML file path.

        """
        salt = datetime.now().strftime('%Y%m%d%H%M%f')
        input_xml = os.path.join(self.data_folder, self.name + '_in_%s.xml' % salt)
        output_xml = os.path.join(self.data_folder, self.name + '_out_%s.xml' % salt)
        partials_xml = os.path.join(self.data_folder, self.name + '_partials_%s.xml' % salt)

        return input_xml, output_xml, partials_xml

    def write_input_file(self, file, inputs):
        # type: (Union[str, etree._ElementTree], Vector) -> None
        """Write the current input values to an input XML file.

        Parameters
        ----------
            file : str or :obj:`etree._ElementTree`
                Path to or :obj:`etree._ElementTree` of an input XML file.

            inputs : Vector
                Input vector of this `Component`.
        """
        # Create new root element and an ElementTree
        root = etree.Element(param_to_xpath(self.inputs_from_xml.keys()[0]).split('/')[1])
        doc = etree.ElementTree(root)

        # Convert all XML param names to XPaths and add new elements to the tree correspondingly
        for param in self.inputs_from_xml:
            if param in inputs:
                xml_safe_create_element(doc, param_to_xpath(param), inputs[param])

        # Write the tree to an XML file
        doc.write(file, pretty_print=True, xml_declaration=True, encoding='utf-8')

    def read_outputs_file(self, file, outputs):
        # type: (Union[str, etree._ElementTree], Vector) -> None
        """Read the outputs from a given XML file and store them in this `Component`'s variables.

        Parameters
        ----------
            file : str or :obj:`etree._ElementTree`
                Path to or :obj:`etree._ElementTree` of an output XML file.

            outputs : Vector
                Output vector of this `Component`.
        """
        # Extract the results from the output xml
        for xpath, value in xml_to_dict(file).items():
            name = xpath_to_param(xpath)
            if name in self.outputs_from_xml and name in outputs:
                outputs[name] = value

    def read_partials_file(self, file, partials):
        # type: (Union[str, etree._ElementTree], Vector) -> None
        """Read the partials from a given XML file and store them in this `Component`'s variables.

        Parameters
        ----------
            file : str or :obj:`etree._ElementTree`
                Path to or :obj:`etree._ElementTree` of a partials XML file.

            partials : Vector
                Partials vector of this `Component`.

        """
        _partials = Partials(file)
        for src, partial in _partials.get_partials().items():
            for tgt, val in partial.items():
                if (src, tgt) in partials:
                    try:
                        partials[src, tgt] = val
                    except Exception as e:
                        print(e.message)

    def compute(self, inputs, outputs):
        # type: (Vector, Vector) -> None
        """Write the input XML file, call `execute()`, and read the output XML file to obtain the results.

        Parameters
        ----------
            inputs : `Vector`
                Input parameters.

            outputs : `Vector`
                Output parameters.
        """

        input_xml, output_xml, _ = self.generate_file_names()

        if self.inputs_from_xml:
            self.write_input_file(input_xml, inputs)
            if self.base_file is not None:
                xml_merge(self.base_file, input_xml)

        # Call execute
        if self.base_file is not None:
            self.execute(self.base_file, output_xml)
            xml_merge(self.base_file, output_xml)
        else:
            self.execute(input_xml, output_xml)

        # If files should not be kept, delete the input XML file
        if not self.keep_files:
            try:
                os.remove(input_xml)
            except OSError:
                pass

        if self.outputs_from_xml:
            self.read_outputs_file(output_xml, outputs)

            # If files should not be kept, delete the output XML file
            if not self.keep_files:
                try:
                    os.remove(output_xml)
                except OSError:
                    pass

    def compute_partials(self, inputs, partials):
        # type: (Vector, Vector) -> None
        """Write the input XML file, call `linearize()`, and read the sensitivities from the resulting XML file.

        Parameters
        ----------
            inputs : `Vector`
                Input parameters.

            partials: `Vector`
                Partials.
        """
        if self.partials_from_xml:
            input_xml, _, partials_xml = self.generate_file_names()

            self.write_input_file(input_xml, inputs)
            self.linearize(input_xml, partials_xml)

            if not self.keep_files:
                try:
                    os.remove(input_xml)
                except OSError:
                    pass

            self.read_partials_file(partials_xml, partials)

            if not self.keep_files:
                try:
                    os.remove(partials_xml)
                except OSError:
                    pass

    def xml_params_as_indep_vars(self, group, params, values, aliases=None):
        # type: (Group, List[str], Union[np.ndarray, Iterable], Optional[List[str]]) -> None
        """Create `IndepVarComp`s for given input params of this `XMLComponent`.

        Parameters
        ----------
            group : :obj:`Group`
                `Group` to add the `IndepVarComp`s to.

            params : list of str
                List of param names. These need to exist in this `XMLComponent`.

            values : :obj:`np.ndarray` or list of numbers
                List of (initial) values for all `IndepVarComp`s.

            aliases : list of str, optional
                List of aliases (promoted names) to give the `IndepVarComp`s.
        """
        if len(params) != len(values) or (aliases is None and len(params) != len(aliases)):
            raise ValueError('number of params, values and optionally aliases needs to be the same')

        for param in params:
            if param not in self.inputs_from_xml:
                raise ValueError('at least one param given is not a param of this XMLComponent (%s)' % param)

        for index, param in enumerate(params):
            if aliases is None:
                alias = 'INDEP_' + param_to_xpath(param).split('/')[-1].split('[')[0]
            else:
                alias = aliases[index]

            group.add(alias, IndepVarComp(alias, val=values[index]), promotes=[alias])
            group.connect(alias, param)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright 2018 D. de Vries

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This file contains the definition of general utility functions.
"""
from __future__ import absolute_import, division, print_function

import re
import warnings

import numpy as np
from lxml import etree
from openmdao.core.driver import Driver
from typing import Callable, Any, Optional, Union, Type


def try_hard(fun, *args, **kwargs):
    # type: (Callable, *Any, **Any) -> Any
    """Try repeatedly to call a function until it returns successfully.

    Utility function that repeatedly tries to call a given function with the given arguments until that function
    successfully returns. It is possible to limit the maximum number of attempts by setting the try_hard_limit argument.


    Parameters
    ----------
        fun : function
            The function to try to call.

        *args
            Any ordered arguments to pass to the function.

        **kwargs
            Any named arguments to pass to the function.


    Returns
    -------
        any
            The return value of the function to be called.

    Notes
    -----
        This function is part of a workaround to deal with APIs crashing in a non-predictable manner, seemingly random
        way. It was found that, when simply trying to call such functions again, the problem seemed to not exist
        anymore.
    """
    warnings.simplefilter('always', UserWarning)
    try_hard_limit = -1
    if kwargs is not None:
        if 'try_hard_limit' in kwargs.keys():
            try_hard_limit = kwargs['try_hard_limit']
            del kwargs['try_hard_limit']

    msg = None
    return_value = None
    successful = False
    attempts = 0
    while not successful:
        try:
            return_value = fun(*args, **kwargs)
            successful = True
        except Exception as e:
            attempts += 1

            if msg is None:
                msg = 'Had to try again to evaluate: %s(' % fun.__name__ + ', '.join(['%s' % arg for arg in args])
                if kwargs is not None:
                    msg += ', '.join(['%s=%s' % (key, value) for key, value in kwargs.items()])
                msg += '). The following exception was raised: "%s"' % e.message

            if 0 < try_hard_limit <= attempts:
                raise
            else:
                warnings.warn(msg)

    return return_value


class CachedProperty(property):
    """Subclass of `property` using a cache to avoid recalculating an expensive `property` every time it is read.

    An attribute can be decorated with this class is the same way as with a normal `property`. It adds the possibility
    to invalidate the cache when necessary.
    """

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        # type: (Optional[Callable], Optional[Callable], Optional[Callable], Optional[str]) -> None
        """Initialize the `CachedProperty`.

        Parameters
        ----------
            fget, fset, fdel : function, optional
                Getter, setter, and deleter functions.

            doc : str, optional
                Docstring of the property.
        """
        super(CachedProperty, self).__init__(fget, fset, fdel, doc)
        self.__cache = None
        self.__dirty = True

    def __get__(self, instance, owner=None):
        # type: (Any, Optional[type]) -> Any
        """Get the value of the property.

        This method checks if the cache of this property is still valid first. If it is, it simply returns the cached
        value. If it isn't, it calls the `super()` to recompute the cached variable, stores it, and then returns the
        newly calculated value.

        Parameters
        ----------
            instance : any
                The instance through which the attribute is accessed.

            owner : type, optional
                The owner of the attribute.

        Returns
        -------
            any
                The value of the attribute.
        """
        if self.__dirty:
            self.__cache = super(CachedProperty, self).__get__(instance, owner)
            self.__dirty = False
        return self.__cache

    def invalidate(self):
        # type: () -> None
        """Marks the cache of the property as invalid, prompting its recomputation the next time it is accessed."""
        self.__dirty = True


def parse_string(s):
    # type: (str) -> Union[str, np.ndarray, float]
    """Convert a string to a numpy array of floats or float if possible.

    The string is returned unchanged if it cannot be converted to a numpy array of floats or float.

    Parameters
    ----------
        s : str
            String to be converted.

    Returns
    -------
        str or np.ndarray or float
            Parsed string or the string itself.
    """
    v = re.sub(r'[\[\]]', '', s)

    if ',' in v:
        v = v.split(',')
    elif ';' in v:
        v = v.split(';')

    try:
        v = np.atleast_1d(np.array(v, dtype=float))
        if v.size == 1:
            v = v[0]
        return v
    except ValueError:
        return s


def parse_cmdows_value(elem):
    # type: (etree._Element) -> Union[str, np.ndarray, float]
    """Convert an XML element from a CMDOWS file to a value.

    Parameters
    ----------
        elem : :obj:`_Element`
            `etree._Element` to convert.

    Returns
    -------
        str or np.ndarray or float
            Converted element.
    """
    if len(list(elem)) > 1:
        return np.array([parse_string(child.text) for child in elem])
    else:
        return parse_string(elem.text)


def normalized_to_bounds(driver):
    # type: (Type[Driver]) -> Type[NormalizedDriver]
    """Decorate a `Driver` to adjust its ``adder``/``scaler`` attributes normalizing the ``desvar``s.

    This decorator automatically adjusts the adder and scalar attributes of the design variables belonging to the
    targeted ``OpenMDAO`` `Driver` class such that the design variables are normalized to their bounds.

    Parameters
    ----------
        driver : :obj:`Driver`
            `Driver` to normalize the design variables of.

    Returns
    -------
        :obj:`NormalizedDriver`
            Instance of `NormalizedDriver` which inherits from the given `Driver`.

    Examples
    --------
        @normalized_to_bounds\n
        class MyNormalizedDriver(Driver):
            # My design variables will now automatically be normalized to their bounds.
            pass
    """

    class NormalizedDriver(driver):
        """Wrapper class for the `normalized_to_bounds` decorator.

        This class adds a static function to the `Driver` it inherits from, which will intercept all `add_desvar()`
        calls to the wrapped `Driver` class to change its ``adder``/``scaler`` attributes depending on the given
        upper and lower bounds.
        """

        @staticmethod
        def normalize_to_bounds(func):
            # type: (Callable) -> Callable
            """Wrap the function handle of the `add_desvar()` function.

            Parameters
            ----------
                func : function
                    Function handle of the `add_desvar()` function.

            Returns
            -------
                func : function
                    Function wrapping the `add_desvar()` function, which adds logic to calculate ``adder``/``scaler``.
            """

            def new_function(name,          # type: str
                             lower=None,    # type: Optional[Union[float, np.ndarray]]
                             upper=None,    # type: Optional[Union[float, np.ndarray]]
                             *args, **kwargs):
                # type: (...) -> None
                """Wrap the `add_desvar()` function call.

                Inner wrapper function which will set the ``adder`` and ``scaler`` `kwargs` of the wrapped
                `add_desvar()` method before calling it.

                Parameters
                ----------
                    name : str
                        Name of the design variable to add.

                    lower : float or list of float, optional
                        Lower bound(s) of the design variable.

                    upper : float or list of float, optional
                        Upper bound(s) of the design variable.

                    *args
                        Any extra, ordered arguments to pass to the `add_desvar()` method.

                    **kwargs
                        Any extra, named arguments to pass to the `add_desvar()` method.
                """
                if lower is not None:
                    adder = -lower
                else:
                    adder = 0.

                if upper is not None:
                    scaler = 1./(upper + adder)
                else:
                    scaler = 1.

                if len(args) > 4:
                    args = args[:-1]
                elif len(args) > 3:
                    args = args[:-1]

                if 'adder' in kwargs:
                    del kwargs['adder']
                if 'scaler' in kwargs:
                    del kwargs['scaler']

                func(name, lower, upper, adder=adder, scaler=scaler, *args, **kwargs)

            return new_function

        def __getattribute__(self, item):
            """Intercept any calls to the `add_desvar()` method of the Driver class.

            This ``hook`` checks if `add_desvar()` is called. If so, it returns the wrapped function instead of the
            clean `add_desvar()` call.

            Parameters
            ----------
                item : str
                    Name of the attribute.

            Returns
            -------
                any
                    The attribute that was requested or the wrapped call to `add_desvar()` if it is requested.
            """
            x = super(NormalizedDriver, self).__getattribute__(item)
            if item in ['add_desvar']:
                return self.normalize_to_bounds(x)
            else:
                return x

    return NormalizedDriver


re_sys_name_char = re.compile(r'[^_a-zA-Z0-9]')
re_sys_name_starts = re.compile(r'^[a-zA-Z]')


def str_to_valid_sys_name(string):
    # type: (str) -> str
    """Ensures a given string is a valid OpenMDAO system name."""
    sys_name = re_sys_name_char.sub('', string)
    while not re_sys_name_starts.match(sys_name):
        sys_name = sys_name[1:]
    return sys_name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright 2018 D. de Vries

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This file contains the definition of the AbstractDiscipline interface class.
"""
from __future__ import absolute_import, division, print_function

import abc
import inspect
import json
import os

from openlego.partials.partials import Partials


class AbstractDiscipline(object):
    """Defines the common interface for all disciplines within ``OpenLEGO``."""
    __metaclass__ = abc.ABCMeta

    @property
    def name(self):
        # type: () -> str
        """:obj:`str`: Name of this discipline."""
        return self.__class__.__name__

    @property
    def version(self):
        # type: () -> float
        """:obj:`str`: Version number of this discipline."""
        return 1.0

    @property
    def creator(self):
        # type: () -> str
        """:obj:`str`: Name of the person that created this discipline."""
        return 'Placeholder'

    @property
    def description(self):
        # type: () -> str
        """:obj:`str`: Description of this discipline."""
        return 'Abstract discipline'

    @property
    def precision(self):
        # type: () -> int
        """:obj:`int`: Precision of this discipline."""
        return 0

    @property
    def path(self):
        # type: () -> str
        """:obj:`str`: Path at which this discipline resides."""
        return os.path.dirname(inspect.getfile(self.__class__))

    @property
    def in_file(self):
        # type: () -> str
        """:obj:`str`: Path of the template input XML file of this discipline."""
        return os.path.join(self.path, self.name + '-input.xml')

    @property
    def out_file(self):
        # type: () -> str
        """:obj:`str`: Path of the template output XML file of this discipline."""
        return os.path.join(self.path, self.name + '-output.xml')

    @property
    def json_file(self):
        # type: () -> str
        """:obj:`str`: Path of the information JSON file of this discipline."""
        return os.path.join(self.path, self.name + '-info.json')

    @property
    def partials_file(self):
        # type: () -> str
        """:obj:`str`: Path of the partials XML file of this discipline."""
        return os.path.join(self.path, self.name + '-partials.xml')

    @property
    def supplies_partials(self):
        # type: () -> bool
        """Set to True to indicate this discipline supplies gradients."""
        return False

    @abc.abstractmethod
    def generate_input_xml(self):
        # type: () -> str
        """Generate the template input XML for this discipline.

        This method should be implemented to define the input template of a specific discipline.

        Returns
        -------
            str
                String representation of the template input XML.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def generate_output_xml(self):
        # type: () -> str
        """Generate the template output XML for this discipline.

        This method should be implemented to define the output template of a specific discipline.

        Returns
        -------
            str
                String representation of the template output XML file.
        """
        raise NotImplementedError

    def generate_info_json(self):
        # type: () -> str
        """Generate the information JSON file for this discipline.

        This method should be overridden or extended to specify a non-standard info JSON file for a specific discipline.

        Returns
        -------
            str
                String representation of the info JSON file
        """
        return json.dumps({'general_info': {'name': self.name,
                                            'version': self.version,
                                            'creator': self.creator,
                                            'description': self.description},
                           'execution_info': [{'mode': 'main',
                                               'description': 'main execution mode',
                                               'precision': self.precision}]}, indent=4)

    def generate_partials_xml(self):
        # type: () -> str
        """Generate the template partials XML file for this discipline.

        This method should be implemented to define for which inputs this discipline can provide the sensitivities.

        Returns
        -------
            str
                String representation of the template partials XML file.
        """
        return Partials().get_string()

    def deploy(self):
        # type: () -> None
        """Deploy this discipline's template in-/output, partials XML files and its information JSON file."""
        with open(self.in_file, 'w') as f:
            f.write(self.generate_input_xml())
        with open(self.out_file, 'w') as f:
            f.write(self.generate_output_xml())
        with open(self.json_file, 'w') as f:
            f.write(self.generate_info_json())
        with open(self.partials_file, 'w') as f:
            f.write(self.generate_partials_xml())

    @staticmethod
    @abc.abstractmethod
    def execute(in_file, out_file):
        # type: (str, str) -> None
        """Execute this discipline with the given in- and output XML files.

        This method should be implemented to define the execution of a specific discipline.

        Parameters
        ----------
            in_file : str
                Path to the input XML file.

            out_file : str
                Path to the output XML file.
        """
        raise NotImplementedError

    @staticmethod
    def linearize(in_file, partials_file):
        # type: (str, str) -> None
        """Compute the sensitivities of a given input XML file and write them to a given partials XML file.

        This method should be implemented to define the linearization of a specific discipline. By default a discipline
        is considered a 'black box', and no sensitivities are provided.

        Parameters
        ----------
            in_file : str
                Path to the input XML file.

            partials_file : str
                Path to the sensitivities XML file.
        """
        Partials().write(partials_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright 2018 D. de Vries

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This file contains the definition of the `LEGOModel` class.
"""
from __future__ import absolute_import, division, print_function

import imp
import os
import warnings

import numpy as np
from lxml import etree
from lxml.etree import _Element, _ElementTree
from openmdao.api import Group, IndepVarComp, LinearBlockGS, NonlinearBlockGS, LinearBlockJac, NonlinearBlockJac, \
    LinearRunOnce, ExecComp, NonlinearRunOnce, DirectSolver
from typing import Union, Optional, List, Any, Dict, Tuple

from openlego.utils.general_utils import CachedProperty, parse_cmdows_value, str_to_valid_sys_name, parse_string
from openlego.utils.xml_utils import xpath_to_param, xml_to_dict
from openlego.utils.cmdows_utils import get_element_by_uid, get_related_parameter_uid, get_loop_nesting_obj
from .abstract_discipline import AbstractDiscipline
from .discipline_component import DisciplineComponent


class InvalidCMDOWSFileError(ValueError):

    def __init__(self, reason=None):
        msg = 'Invalid CMDOWS file'
        if reason is not None:
            msg += ': {}'.format(reason)
        super(InvalidCMDOWSFileError, self).__init__(msg)


class LEGOModel(Group):
    """Specialized OpenMDAO Group class representing the model specified by a CMDOWS file.

    An important note about this class in the context of OpenMDAO is that the aggregation pattern of the root Group
    class the base Problem class has is changed into a stronger composition pattern. This is because this class directly
    controls the creation and assembly of this class by making use of Python's @property decorator. It is not possible,
    nor should it be attempted, to manually inject a different instance of Group  in place of these, because the
    correspondence between the CMDOWS file and the Problem can then no longer be guaranteed.

    Attributes
    ----------
        cmdows_path
        kb_path
        discipline_components
        block_order
        coupled_blocks
        system_order
        system_variables
        system_inputs
        driver
        coordinator

        data_folder : str, optional
            Path to the folder in which to store all data generated during the `Problem`'s execution.

        base_xml_file : str, optional
            Path to an XML file which should be kept up-to-date with the latest data describing the problem.
    """

    def __init__(self, cmdows_path=None, kb_path='', data_folder=None, base_xml_file=None, **kwargs):
        # type: (Optional[str], Optional[str], Optional[str], Optional[str]) -> None
        """Initialize a CMDOWS Problem from a given CMDOWS file and knowledge base.

        It is also possible to specify where (temporary) data should be stored, and if a base XML
        file should be kept up-to-data.

        Parameters
        ----------
        cmdows_path : str, optional
            Path to the CMDOWS file.

        kb_path : str, optional
            Path to the knowledge base.

        data_folder : str, optional
            Path to the data folder in which to store all files and output from the problem.

        base_xml_file : str, optional
            Path to a base XML file to update with the problem data.
        """
        self._cmdows_path = cmdows_path
        self._kb_path = kb_path
        self.data_folder = data_folder
        self.base_xml_file = base_xml_file

        super(LEGOModel, self).__init__(**kwargs)
        self.linear_solver = LinearRunOnce()
        self.nonlinear_solver = NonlinearRunOnce()

    def __getattribute__(self, name):
        # type: (str) -> Any
        """Check the integrity before returning any of the cached variables.

        Parameters
        ----------
        name : str
            Name of the attribute to read.

        Returns
        -------
            any
                The value of the requested attribute.
        """
        if name != '__class__' and name != '__dict__':
            if name in [_name for _name, value in self.__class__.__dict__.items() if isinstance(value, CachedProperty)]:
                self.__integrity_check()
        return super(LEGOModel, self).__getattribute__(name)

    def __setattr__(self, name, value):
        # type: (str, Any) -> None
        """Bypass setting coordinator and coupled_group attributes.

        Parameters
        ----------
            name : str
                Name of the attribute.

            value : any
                Value to set the attribute to.
        """
        if name not in ['coordinator', 'coupled_group']:
            super(LEGOModel, self).__setattr__(name, value)

    def __integrity_check(self):
        # type: () -> None
        """Ensure a CMDOWS file has been supplied.

        Raises
        ------
            ValueError
                If no CMDOWS file has been supplied
        """
        if self._cmdows_path is None:
            raise ValueError('No CMDOWS file specified!')

    def invalidate(self):
        # type: () -> None
        """Invalidate the instance.

        All computed (cached) properties will be recomputed upon being read once the instance has been invalidated."""
        for value in self.__class__.__dict__.values():
            if isinstance(value, CachedProperty):
                value.invalidate()

    def does_value_fit(self, name, val):
        # type: (str, Union[str, float, np.ndarray]) -> bool
        """Check whether a given value has the correct size to be assigned to a given variable.

        Parameters
        ----------
            name : str
                Name of the variable.

            val : str or float or np.ndarray
                Value to check.

        Returns
        -------
            bool
                `True` if the value fits, `False` if not.
        """
        return (isinstance(val, np.ndarray) and val.size == self.variable_sizes[name]) \
            or (not isinstance(val, np.ndarray) and self.variable_sizes[name] == 1)

    @property
    def cmdows_path(self):
        # type: () -> str
        """:obj:`str`: Path to the CMDOWS file this class corresponds to.

        When this property is set the instance is automatically invalidated.
        """
        return self._cmdows_path

    @cmdows_path.setter
    def cmdows_path(self, cmdows_path):
        # type: (str) -> None
        self._cmdows_path = cmdows_path
        self.invalidate()

    @property
    def kb_path(self):
        # type: () -> str
        """:obj:`str`: Path to the knowledge base.

        When this property is set the instance is automatically invalidated.
        """
        return self._kb_path

    @kb_path.setter
    def kb_path(self, kb_path):
        # type: (str) -> None
        self._kb_path = kb_path
        self.invalidate()

    @CachedProperty
    def elem_cmdows(self):
        # type: () -> _Element
        """:obj:`etree._Element`: Root element of the CMDOWS XML file."""
        return etree.parse(self.cmdows_path).getroot()

    @CachedProperty
    def elem_problem_def(self):
        # type: () -> _Element
        """:obj:`etree._Element`: The problemDefition element of the CMDOWS file."""
        return self.elem_cmdows.find('problemDefinition')

    @CachedProperty
    def elem_workflow(self):
        # type: () -> _Element
        """:obj:`etree._Element`: The workflow element of the CMDOWS file."""
        return self.elem_cmdows.find('workflow')

    @CachedProperty
    def elem_params(self):
        # type: () -> _Element
        """:obj:`etree._Element`: The problemRoles/parameters element of the CMDOWS file."""
        params = self.elem_cmdows.find('problemDefinition/problemRoles/parameters')
        if params is None:
            raise InvalidCMDOWSFileError('does not contain (valid) parameters in the problemRoles')
        return params

    @CachedProperty
    def elem_arch_elems(self):
        # type: () -> _Element
        """:obj:`etree._Element`: The architectureElements element of the CMDOWS file."""
        arch_elems = self.elem_cmdows.find('architectureElements')
        if arch_elems is None:
            raise InvalidCMDOWSFileError('does not contain (valid) architecture elements')
        return arch_elems

    @CachedProperty
    def elem_loop_nesting(self):
        # type: () -> _Element
        """:obj:`etree._Element`: The loopNesting element of the CMDOWS file."""
        loopnesting_elem = self.elem_workflow.find('processGraph/metadata/loopNesting')
        if loopnesting_elem is None:
            raise InvalidCMDOWSFileError('does not contain loopNesting element')
        return loopnesting_elem

    @CachedProperty
    def has_optimizer(self):
        # type: () -> bool
        """:obj:`bool`: True if there is an optimizer, False if not."""
        if self.elem_arch_elems.find('executableBlocks/optimizers/optimizer') is not None:
            return True
        return False

    @CachedProperty
    def has_doe(self):
        # type: () -> bool
        """:obj:`bool`: True if there is a DOE component, False if not."""
        if self.elem_arch_elems.find('executableBlocks/does/doe') is not None:
            return True
        return False

    @CachedProperty
    def has_driver(self):
        # type: () -> bool
        """:obj:`bool`: True if there is a driver component (DOE or optimizer), False if not."""
        if self.has_doe or self.has_optimizer:
            return True
        return False

    @CachedProperty
    def has_converger(self):
        # type: () -> bool
        """:obj:`bool`: True if there is a converger, False if not."""
        if self.elem_arch_elems.find('executableBlocks/convergers/converger') is not None:
            return True
        return False

    @CachedProperty
    def objective_required(self):
        # type: () -> bool
        """:obj:`bool`: True if an objective value is required, False if not."""
        if self.has_optimizer:
            return True
        return False

    @CachedProperty
    def discipline_components(self):
        # type: () -> Dict[str, DisciplineComponent]
        """:obj:`dict`: Dictionary of discipline components by their design competence ``uID`` from CMDOWS.

        Raises
        ------
            RuntimeError
                If a ``designCompetence`` specified in the CMDOWS file does not correspond to an `AbstractDiscipline`.
        """
        # Ensure that the knowledge base path is specified.
        _discipline_components = dict()
        for design_competence in self.elem_cmdows.iter('designCompetence'):
            if not self._kb_path or not os.path.isdir(self._kb_path):
                raise ValueError('No valid knowledge base path ({}) specified while the CMDOWS file contains design'
                                 ' competences.'.format(self._kb_path))
            uid = design_competence.attrib['uID']
            name = design_competence.find('ID').text
            try:
                fp, pathname, description = imp.find_module(name, [self.kb_path])
                mod = imp.load_module(name, fp, pathname, description)
                cls = getattr(mod, name)  # type: AbstractDiscipline.__class__
                if not issubclass(cls, AbstractDiscipline):
                    raise RuntimeError
            except Exception:
                raise ValueError(
                    'Unable to process CMDOWS file: no proper discipline found for design competence with name %s'
                    % name)
            finally:
                if 'fp' in locals():
                    fp.close()

            component = DisciplineComponent(cls(), data_folder=self.data_folder, base_file=self.base_xml_file)
            _discipline_components.update({uid: component})
        return _discipline_components

    @CachedProperty
    def mapped_parameters(self):
        # type: () -> Dict[str, str]
        """:obj:`dict`: Dictionary of parameters that are mapped in the CMDOWS file, for example as copies."""
        mapped_params = dict()
        for elem_category in self.elem_arch_elems.find('parameters').iterchildren():
            for elem_param in elem_category.iterchildren():
                param, mapped = get_related_parameter_uid(elem_param, self.elem_cmdows)
                mapped_params.update({param: mapped})
        return mapped_params

    @CachedProperty
    def mathematical_functions_inputs(self):
        # type: () -> Dict[str, List[Tuple[str]]]
        """:obj:`dict`: Dictionary of all mathematical function blocks with a list of their input variables."""
        _inputs = dict()
        for mathematical_function in self.elem_cmdows.iter('mathematicalFunction'):
            uid = mathematical_function.attrib['uID']

            local_inputs = list()
            for _input in mathematical_function.iter('input'):
                input_name = xpath_to_param(_input.find('parameterUID').text)
                eq_label = _input.find('equationLabel').text
                local_inputs.append((eq_label, input_name))

            _inputs.update({uid: local_inputs})
        return _inputs

    @CachedProperty
    def mathematical_functions_outputs(self):
        # type: () -> Dict[str, List[Tuple[str]]]
        """:obj:`dict`: Dictionary of all mathematical function blocks with a list of their output variables."""
        _outputs = dict()
        for mathematical_function in self.elem_cmdows.iter('mathematicalFunction'):
            uid = mathematical_function.attrib['uID']

            local_outputs = list()
            for _output in mathematical_function.iter('output'):
                output_name = xpath_to_param(_output.find('parameterUID').text)
                local_outputs.append(output_name)

            _outputs.update({uid: local_outputs})
        return _outputs

    @CachedProperty
    def mathematical_functions_groups(self):
        # type: () -> Dict[str, Group]
        """:obj:`dict`: Dictionary of execute components by their mathematical function ``uID`` from CMDOWS.
        """
        _mathematical_functions = dict()
        for mathematical_function in self.elem_cmdows.iter('mathematicalFunction'):
            uid = mathematical_function.attrib['uID']
            group = Group()

            eq_mapping = dict()
            for output in mathematical_function.iter('output'):
                if output.find('equations') is not None:
                    for equation in output.iter('equation'):
                        if equation.attrib['language'] == 'Python':
                            eq_uid = equation.getparent().attrib['uID']
                            eq_expr = equation.text
                            eq_output = xpath_to_param(output.find('parameterUID').text)

                            promotes = list()
                            for eq_label, input_name in self.mathematical_functions_inputs[uid]:
                                # TODO: This mapping of the input name should get a more sophisticated logic
                                if input_name in self.mapped_parameters and input_name not in self.design_vars:
                                    input_name = self.mapped_parameters[input_name]

                                if eq_label in eq_expr:
                                    promotes.append((eq_label, input_name))

                            eq_mapping.update({eq_uid: eq_output})
                            group.add_subsystem(str_to_valid_sys_name(eq_uid),
                                                ExecComp('output = ' + eq_expr),
                                                promotes=promotes + [('output', eq_output), ])

            extra_output_counter = 0
            for output in mathematical_function.iter('output'):
                eq_uid = output.find('equationsUID')
                if eq_uid is not None:
                    eq_uid = eq_uid.text
                    eq_output = xpath_to_param(output.find('parameterUID').text)

                    group.add_subsystem('extra_output_{}'.format(extra_output_counter),
                                        ExecComp('output = input'),
                                        promotes=[('output', eq_output), ('input', eq_mapping[eq_uid])])
                    extra_output_counter += 1

            _mathematical_functions.update({uid: group})

        return _mathematical_functions

    @CachedProperty
    def variable_sizes(self):
        # type: () -> Dict[str, int]
        """:obj:`dict`: Dictionary of the sizes of all variables by their names."""
        variable_sizes = {}
        for component in self.discipline_components.values():
            for name, value in component.variables_from_xml.items():
                variable_sizes.update({name: np.atleast_1d(value).size})

        for local_inputs in self.mathematical_functions_inputs.values():
            for _input in local_inputs:
                variable_sizes.update({_input[1]: np.atleast_1d([0]).size})

        for local_outputs in self.mathematical_functions_outputs.values():
            for _output in local_outputs:
                variable_sizes.update({_output: np.atleast_1d([0]).size})

        return variable_sizes

    @CachedProperty
    def coupling_vars(self):
        # type: () -> Dict[str, Dict[str, str]]
        """:obj:`dict`: Dictionary with coupling variables."""
        coupling_vars = dict()

        # First create a map between related param and coupling copy var
        for var in self.elem_arch_elems.iter('couplingCopyVariable'):
            param, mapped = get_related_parameter_uid(var, self.elem_cmdows)
            coupling_vars.update({xpath_to_param(mapped): xpath_to_param(param)})

        # Then update dict with corresponding consistency constraint var
        for convar in self.elem_arch_elems.iter('consistencyConstraintVariable'):
            param = xpath_to_param(convar.find('relatedParameterUID').text)
            if param not in coupling_vars:
                raise InvalidCMDOWSFileError()

            coupling_vars.update({param: {'copy': coupling_vars[param], 'con': xpath_to_param(convar.attrib['uID'])}})
        return coupling_vars

    @CachedProperty
    def coupling_var_copies(self):
        # type: () -> Dict[str, str]
        """:obj:`dict`: Dictionary with coupling variable copies."""
        coupling_var_copies = dict()
        for var, value in self.coupling_vars.items():
            coupling_var_copies.update({var: value['copy']})
        return coupling_var_copies

    @CachedProperty
    def coupling_var_cons(self):
        # type: () -> Dict[str, str]
        """:obj:`dict`: Dictionary with coupling variable constraints."""
        coupling_var_cons = dict()
        for var, value in self.coupling_vars.items():
            if isinstance(value, dict):
                if 'con' in value:
                    coupling_var_cons.update({var: value['con']})
        if coupling_var_cons:
            return coupling_var_cons
        else:
            return None

    @CachedProperty
    def block_order(self):
        # type: () -> List[str]
        """:obj:`list` of :obj:`str`: List of executable block ``uIDs`` in the order specified in the CMDOWS file."""
        positions = list()
        uids = list()
        for block in self.elem_workflow.iterfind('processGraph/metadata/executableBlocksOrder/executableBlock'):
            uid = block.text
            positions.append(int(block.attrib['position']))
            uids.append(uid)
        return [uid for position, uid in sorted(zip(positions, uids))]

    @CachedProperty
    def coupled_blocks(self):
        # type: () -> List[str]
        """:obj:`list` of :obj:`str`: List of ``uIDs`` of the coupled executable blocks specified in the CMDOWS file."""
        _coupled_blocks = []
        for block in self.elem_arch_elems.iterfind(
                'executableBlocks/coupledAnalyses/coupledAnalysis/relatedExecutableBlockUID'):
            _coupled_blocks.append(block.text)
        return _coupled_blocks

    @CachedProperty
    def loop_nesting_dict(self):
        # type: () -> Dict[str, dict]
        """:obj:`dict`: Dictionary of the loopNesting XML element."""
        return get_loop_nesting_obj(self.elem_loop_nesting)

    @CachedProperty
    def loop_element_details(self):
        # type: () -> Dict[str]
        """:obj:`dict` of :obj:`str`: Dictionary with mapping of loop elements specified in the CMDOWS file."""
        _loopelement_details = {}
        for elem in self.elem_arch_elems.iterfind('executableBlocks/coordinators/coordinator'):
            _loopelement_details[elem.attrib['uID']] = 'coordinator'
        for elem in self.elem_arch_elems.iterfind('executableBlocks/convergers/converger'):
            _loopelement_details[elem.attrib['uID']] = 'converger'
        for elem in self.elem_arch_elems.iterfind('executableBlocks/optimizers/optimizer'):
            _loopelement_details[elem.attrib['uID']] = 'optimizer'
        for elem in self.elem_arch_elems.iterfind('executableBlocks/does/doe'):
            _loopelement_details[elem.attrib['uID']] = 'doe'
        return _loopelement_details

    @CachedProperty
    def coupled_hierarchy(self):
        # type: () -> List[dict]
        """:obj:`list`: List containing the hierarchy of the coupled blocks for grouped convergence."""
        return self._get_coupled_hierarchy(self.loop_nesting_dict)

    def _get_coupled_hierarchy(self, hierarchy):
        # type: (List) -> List[dict]
        """:obj:`list`: List containing the hierarchy of the coupled functions which defines the hierarchy of converged
        groups."""
        _coupled_hierarchy = []
        for entry in hierarchy:
            if isinstance(entry, dict):
                keys = entry.keys()
                assert len(keys) == 1, 'One key is expected in the dictionary of a process hierarchy.'
                if self.loop_element_details[keys[0]] == 'converger':
                    _coupled_hierarchy.append(entry)
                else:
                    return self._get_coupled_hierarchy(entry[keys[0]])
        return _coupled_hierarchy

    @CachedProperty
    def system_inputs(self):
        # type: () -> Dict[str, int]
        """:obj:`dict`: Dictionary containing the system input sizes by their names."""
        system_inputs = {}
        for value in self.elem_cmdows.xpath(
                    r'workflow/dataGraph/edges/edge[fromExecutableBlockUID="Coordinator"]/toParameterUID/text()'):
            if 'architectureNodes' not in value or 'designVariables' in value:
                name = xpath_to_param(value)
                system_inputs.update({name: self.variable_sizes[name]})

        return system_inputs

    @CachedProperty
    def design_vars(self):
        # type: () -> Dict[str, Dict[str, Any]]
        """:obj:`dict`: Dictionary containing the design variables' initial values, lower bounds, and upper bounds."""
        desvars = self.elem_params.find('designVariables')
        if desvars is None:
            if self.has_driver:
                raise Exception(
                    'CMDOWS file {} does contain an optimizer, but no (valid) design variables'.format(self.cmdows_path)
                )
            else:
                return {}
        design_vars = {}
        for desvar in desvars:
            name = xpath_to_param(desvar.find('parameterUID').text)

            # Obtain the initial value
            initial = desvar.find('nominalValue')
            if initial is not None:
                initial = parse_cmdows_value(initial)
                if not self.does_value_fit(name, initial):
                    raise ValueError('incompatible size of nominalValue for design variable "%s"' % name)
            else:
                warnings.warn('no nominalValue given for designVariable "%s". Default is all zeros.' % name)
                initial = np.zeros(self.variable_sizes[name])

            # Obtain the lower and upper bounds
            bounds = 2 * [None]  # type: List[Optional[str]]
            limit_range = desvar.find('validRanges/limitRange')
            if limit_range is not None:
                for index, bnd, in enumerate(['minimum', 'maximum']):
                    elem = limit_range.find(bnd)
                    if elem is not None:
                        bounds[index] = parse_cmdows_value(elem)
                        if not self.does_value_fit(name, bounds[index]):
                            raise ValueError('incompatible size of %s for design variable %s' % (bnd, name))
            else:
                bounds[0] = None
                bounds[1] = None

            # Add the design variable to the dict
            node_name = name if name not in self.coupling_vars else self.coupling_vars[name]['copy']
            design_vars.update({node_name: {'initial': initial,
                                            'lower': bounds[0], 'upper': bounds[1],
                                            'ref0': bounds[0], 'ref': bounds[1]}})
        return design_vars

    @CachedProperty
    def constraints(self):
        # type: () -> Dict[str, Dict[str, Any]]
        """:obj:`dict`: Dictionary containing the constraints' lower, upper, and equals reference values."""
        convars = self.elem_params.find('constraintVariables')
        constraints = {}
        if convars is not None:
            for convar in convars:
                con = {'lower': None, 'upper': None, 'equals': None}
                name = xpath_to_param(convar.find('parameterUID').text)

                if self.coupling_var_cons is not None and name in self.coupling_var_cons.values():
                    # If this is a coupling variable consistency constraint, equals should just be zero
                    for key, value in self.coupling_var_cons.items():
                        if name == value:
                            size = self.variable_sizes[key]
                            if size == 1:
                                con['equals'] = 0.
                            else:
                                con['equals'] = np.zeros(self.variable_sizes[key])
                            break
                else:
                    # Obtain the reference value of the constraint
                    constr_ref = convar.find('referenceValue')  # type: etree._Element
                    ref_vals = []
                    if constr_ref is not None:
                        refs_str = constr_ref.text
                        if ';' in refs_str:
                            refs = refs_str.split(';')
                        else:
                            refs = [refs_str]
                        for ref in refs:
                            ref_val = parse_string(ref)
                            if isinstance(ref_val, str):
                                raise ValueError('referenceValue for constraint "%s" is not numerical' % name)
                            elif not self.does_value_fit(name, ref_val):
                                warnings.warn(
                                    'incompatible size of constraint "%s". Will assume the same for all.' % name)
                                ref_val = np.ones(self.variable_sizes[name]) * np.atleast_1d(ref_val)[0]
                            ref_vals.append(ref_val)
                    else:
                        warnings.warn('no referenceValue given for constraint "%s". Default is all zeros.' % name)
                        ref_vals = [np.zeros(self.variable_sizes[name])]

                    # Process the constraint type
                    constr_type = convar.find('constraintType')
                    if constr_type is not None:
                        if constr_type.text == 'inequality':
                            constr_oper = convar.find('constraintOperator')
                            if constr_oper is not None:
                                opers_str = constr_oper.text
                                if ';' in opers_str:
                                    opers = opers_str.split(';')
                                else:
                                    opers = [opers_str]
                                for idx, oper in enumerate(opers):
                                    if oper == '>=' or oper == '>':
                                        con['lower'] = ref_vals[idx]
                                    elif oper == '<=' or oper == '<':
                                        con['upper'] = ref_vals[idx]
                                    else:
                                        raise ValueError(
                                            'invalid constraintOperator "%s" for constraint "%s"' % (oper, name))
                            else:
                                warnings.warn(
                                    'no constraintOperator given for inequality constraint. Default is "&lt;=".')
                                con['upper'] = ref_vals[0]
                        elif constr_type.text == 'equality':
                            if convar.find('constraintOperator') is not None:
                                warnings.warn('constraintOperator given for an equalityConstraint will be ignored')
                            con['equals'] = ref_vals[0]
                        else:
                            raise ValueError(
                                'invalid constraintType "%s" for constraint "%s".' % (constr_type.text, name))
                    else:
                        warnings.warn('no constraintType specified for constraint "%s". Default is a <= inequality.')
                        con['upper'] = ref

                # Add constraint to the dictionary
                constraints.update({name: con})
        return constraints

    @CachedProperty
    def objective(self):
        # type: () -> str
        """:obj:`str`: Name of the objective variable."""
        objvars = self.elem_params.find('objectiveVariables')
        if self.objective_required:
            if objvars is None:
                raise InvalidCMDOWSFileError('does not contain (valid) objective variables')
            if len(objvars) > 1:
                raise InvalidCMDOWSFileError('contains multiple objectives, but this is not supported')
            return xpath_to_param(objvars[0].find('parameterUID').text)
        else:
            pass

    def _configure_coupled_groups(self, hierarchy, root=True):
        # type: (List) -> Optional[Group]
        """:obj:`Group`, optional: Group wrapping the coupled blocks with a converger specified in the CMDOWS file.

        This method enables the iterative configuration of groups of distributed convergers based on the convergence
        hierarchy.
        """
        if not root:
            coupled_group = Group()
        for entry in hierarchy:
            if isinstance(entry, dict):  # if entry specifies a coupled group
                uid = entry.keys()[0]
                if root:
                    subsys = self.add_subsystem(str_to_valid_sys_name(uid),
                                                self._configure_coupled_groups(entry[uid], False), ['*'])
                else:
                    subsys = coupled_group.add_subsystem(str_to_valid_sys_name(uid),
                                                         self._configure_coupled_groups(entry[uid], False), ['*'])
                conv_elem = get_element_by_uid(self.elem_arch_elems, uid)
                # Define linear solver
                linsol_elem = conv_elem.find('settings/linearSolver')
                if isinstance(linsol_elem, _Element):
                    if linsol_elem.find('method').text == 'Gauss-Seidel':
                        linsol = subsys.linear_solver = LinearBlockGS()
                    elif linsol_elem.find('method').text == 'Jacobi':
                        linsol = subsys.linear_solver = LinearBlockJac()
                    else:
                        raise ValueError('Specified convergerType "{}" is not supported.'
                                         .format(linsol_elem.find('method').text))
                    linsol.options['maxiter'] = int(linsol_elem.find('maximumIterations').text)
                    linsol.options['atol'] = float(linsol_elem.find('convergenceToleranceAbsolute').text)
                    linsol.options['rtol'] = float(linsol_elem.find('convergenceToleranceRelative').text)
                else:
                    subsys.linear_solver = DirectSolver()
                    warnings.warn('Linear solver was not defined in CMDOWS file for converger {}. linear_solver set to'
                                  ' default "DirectSolver()".'.format(str_to_valid_sys_name(uid)))

                # Define nonlinear solver
                nonlinsol_elem = conv_elem.find('settings/nonlinearSolver')
                if isinstance(nonlinsol_elem, _Element):
                    if nonlinsol_elem.find('method').text == 'Gauss-Seidel':
                        nonlinsol = subsys.nonlinear_solver = NonlinearBlockGS()
                    elif nonlinsol_elem.find('method').text == 'Jacobi':
                        nonlinsol = subsys.nonlinear_solver = NonlinearBlockJac()
                    else:
                        raise ValueError('Specified convergerType "{}" is not supported.'
                                         .format(nonlinsol_elem.find('method').text))
                    nonlinsol.options['maxiter'] = int(nonlinsol_elem.find('maximumIterations').text)
                    nonlinsol.options['atol'] = float(nonlinsol_elem.find('convergenceToleranceAbsolute').text)
                    nonlinsol.options['rtol'] = float(nonlinsol_elem.find('convergenceToleranceRelative').text)
                else:
                    subsys.nonlinear_solver = NonlinearRunOnce()
                    warnings.warn('Nonlinear solver was not defined in CMDOWS file for converger {}. nonlinear_solver'
                                  ' set to default "NonlinearRunOnce()".'.format(str_to_valid_sys_name(uid)))
            elif isinstance(entry, str):  # if entry specifies an executable block
                if root:
                    raise AssertionError('Code was not expected to get here for root == True.')
                # Get the correct DisciplineComponent or MathematicalFunction
                promotes = ['*']  # type: List[Union[str, Tuple[str, str]]]
                uid = entry
                if uid in self.discipline_components:
                    block = self.discipline_components[uid]
                elif uid in self.mathematical_functions_groups:
                    block = self.mathematical_functions_groups[uid]
                else:
                    raise RuntimeError('uID {} not found in discipline_components, nor mathematical_functions_groups.'
                                       .format(uid))
                # Add the block to the group
                coupled_group.add_subsystem(str_to_valid_sys_name(uid), block, promotes)
            else:
                raise ValueError('Unexpected value type {} encountered in the coupled_hierarchy {}.'
                                 .format(type(entry), hierarchy))
        if root:
            return subsys
        else:
            return coupled_group

    # TODO: implement this property in phase 3 (required for distributed architectures such as BLISS-2000 and CO)
    @CachedProperty
    def subsystem_optimization_groups(self):
        # type: () -> Optional[list]
        """:obj:`list`, optional: list containing groups of suboptimizations used in distributed architectures.

        If not subsystem optimizations are required based on the CMDOWS file then this property is `None`.
        """
        if self.subsystem_optimizations:
            pass
            # Add inputs (standardized?)
            # Add output
            # Declare partials
            # Set subproblem
            # Define copies in params subsystem
            # Define design variables
            # Define components
            # Connect everything together (through promotion?)
            # Set subproblem optimizer
            # Add design variables, objective, constraints
            # Setup and final setup

    @CachedProperty
    def system_order(self):
        # type: () -> List[str]
        """:obj:`list` of :obj:`str`: List system names in the order specified in the CMDOWS file."""
        _system_order = ['coordinator']
        coupled_group_set = False
        n = 0
        for block in self.block_order:
            if block in self.coupled_blocks:
                n += 1
                if not coupled_group_set:
                    for entry in self.coupled_hierarchy:
                        _system_order.append(str_to_valid_sys_name(entry.keys()[0]))
                    coupled_group_set = True
            elif block in self.discipline_components or block in self.mathematical_functions_groups:
                n += 1
                _system_order.append(str_to_valid_sys_name(block))

        if n < len(self.discipline_components) + len(self.mathematical_functions_groups):
            raise InvalidCMDOWSFileError('executableBlocksOrder is incomplete')

        return _system_order

    @CachedProperty
    def coordinator(self):
        # type: () -> IndepVarComp
        """:obj:`IndepVarComp`: An `IndepVarComp` representing the system's ``Coordinator`` block.

        This `IndepVarComp` takes care of all system input parameters and initial values of design variables.
        """
        coordinator = IndepVarComp()

        # Add design variables
        for name, value in self.design_vars.items():
            coordinator.add_output(name, value['initial'])

        # Add system constants
        for name, shape in self.system_inputs.items():
            if name not in self.design_vars.keys():
                coordinator.add_output(name, shape=shape)

        return coordinator

    def setup(self):
        # type: () -> None
        """Assemble the LEGOModel using the the CMDOWS file and knowledge base."""
        # Add the coordinator
        self.add_subsystem('coordinator', self.coordinator, ['*'])

        # Add all pre-coupling and post-coupling components
        for name, component in self.discipline_components.items():
            if name not in self.coupled_blocks:
                promotes = ['*']
                # Change input variable names if they are provided as copies of coupling variables
                for i in component.inputs_from_xml.keys():
                    if i in self.coupling_vars:
                        if isinstance(self.coupling_vars[i], dict):
                            if 'copy' in self.coupling_vars[i]:
                                promotes.append((i, self.coupling_vars[i]['copy']))
                self.add_subsystem(str_to_valid_sys_name(name), component, promotes)
        for name, component in self.mathematical_functions_groups.items():
            if name not in self.coupled_blocks:
                # TODO: Adjust promotion of variables for copies?
                self.add_subsystem(str_to_valid_sys_name(name), component, ['*'])

        # Add the coupled groups
        if self.coupled_hierarchy:
            self._configure_coupled_groups(self.coupled_hierarchy, True)

        # Put the blocks in the correct order
        self.set_order(list(self.system_order))

        # Add the design variables
        for name, value in self.design_vars.items():
            self.add_design_var(name, lower=value['lower'], upper=value['upper'], ref0=value['ref0'], ref=value['ref'])

        # Add the constraints
        for name, value in self.constraints.items():
            self.add_constraint(name, lower=value['lower'], upper=value['upper'], equals=value['equals'])

        # Add the objective
        if self.objective:
            self.add_objective(self.objective)

    def initialize_from_xml(self, xml):
        # type: (Union[str, _ElementTree]) -> None
        """Initialize the problem with initial values from an XML file.

        This function can only be called after the problem's setup method has been called.

        Parameters
        ----------
            xml : str or :obj:`etree._ElementTree`
                Path to an XML file or an instance of `etree._ElementTree` representing it.
        """
        for xpath, value in xml_to_dict(xml).items():
            name = xpath_to_param(xpath)
            if name in self._outputs:
                self._outputs[name] = value
            elif name in self._inputs:
                self._inputs[name] = value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright 2018 D. de Vries

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This file contains the setup code for the OpenLEGO package.
"""
from setuptools import setup

from openlego import __version__ as version


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='openlego',
    version=version,
    description='An Open-source link between the AGILE project and the OpenMDAO framework',
    long_description=readme(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6'
    ],
    keywords='optimization agile multidisciplinary kadmos openmdao engineering xml cpacs',
    url='https://github.com/daniel-de-vries/OpenLEGO',
    download_url='https://github.com/daniel-de-vries/OpenLEGO/tarball/' + version,
    author='Daniël de Vries',
    author_email='danieldevries6@gmail.com',
    license='Apache Software License',
    packages=[
        'openlego',
        'openlego.core',
        'openlego.utils',
        'openlego.docs',
        'openlego.test_suite',
        'openlego.test_suite.test_examples'
    ],
    install_requires=[
        'kadmos==0.8',
        'openmdao',
        'lxml',
        'numpy',
        'matplotlib',
        'ssbj-kadmos'
    ],
    include_package_data=True,
    zip_safe=False
)



from __future__ import absolute_import, division, print_function

# Core
from .core.xml_component import XMLComponent
from .core.discipline_component import DisciplineComponent
from .core.abstract_discipline import AbstractDiscipline
from .core.model import LEGOModel
